#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "fibheap_int.h"

fibheap_t *
fibheap_create(fibheap_cmp cmp, void *data)
{
  fibheap_t *h = (fibheap_t *)calloc(1, sizeof(fibheap_t));
  
  h->n = 0;
  h->min = 0;
  h->cmp = cmp;
  h->neginf = data;
  return h;
}

void
free_recursive(fibheap_node *node, fibheap_free_fcn f)
{
  fibheap_node *tmp;

  if (node->child != 0)
	free_recursive(node->child, f);
  tmp = 0;
  if (node != node->right) {
    tmp = node->right;
    node->left->right = tmp;
    tmp->left = node->left;
  }   
  f(node->key);
  free(node);
  if (tmp != 0)
    free_recursive(tmp, f);
}

void
fibheap_free(fibheap_t *h, fibheap_free_fcn f)
{
  fibheap_node *node;
  
  node = h->min;
  if (node != 0) 
    free_recursive(node, f);
  free(h->neginf);
  free(h);
}

void
concatenate(fibheap_t *h, fibheap_node *node2)
{
  fibheap_node *node1;

  node1 = h->min;
  if (node1 == 0) 
    h->min = node2;
  else {
    node1->right->left = node2->left;
    node2->left->right = node1->right;
    node1->right = node2;
    node2->left = node1;
  }	
}

fibheap_node *
fibheap_insert(fibheap_t *h, void *data)
{
  fibheap_node *node = (fibheap_node *)calloc(1, sizeof(fibheap_node));

  node->key = data;
  node->degree = 0;
  node->p = 0;
  node->child = 0;
  node->left = node;
  node->right = node;
  node->mark = 0;
 
  concatenate(h, node);  
  if (h->cmp(node->key, h->min->key) < 0)
    h->min=node;
  h->n ++;
  return node;
}

void
fibheap_union(fibheap_t *h1, fibheap_t *h2)
{
  concatenate(h1, h2->min);
  if ((h1->min==0) || 
      ((h2->min!=0) && (h1->cmp(h2->min->key, h1->min->key)<0)))
    h1->min=h2->min;
  h1->n = h1->n+h2->n;
  free(h2);
}

void
fib_heap_link(fibheap_t *h, fibheap_node *y, fibheap_node *x)
{
  if (x->child == 0) {
    x->child=y;
    y->p=x;
    //    y->left=y;
    //	  y->right=y;
    //	  x->left=x;
    //	  x->right=x;
  }
  else {
    y->p = x;	  
    x->child->right->left = y;
    y->right = x->child->right;
    x->child->right = y;
    y->left = x->child;
    //	  x->left=x;
    //    x->right=x;
  }  
  x->degree ++;
  y->mark = 0;
}


void
consolidate(fibheap_t *h)
{
  int i, D, d;
  float D1;
  D1 = log(h->n)/log(1.618);
  D = (int)D1;
  fibheap_node *A[50], *x, *x1, *y, *tmp;

  for (i=0; i<=D; i++) 
    A[i]=0;
  x = h->min;
  while (x!=0) {
    if (x == x->right)
      x1 = 0;
    else
      x1 = x->right;
    d = x->degree;
    if (x != x->right) {
      x->left->right = x->right;
      x->right->left = x->left;
      x->left = x;
      x->right = x;
    }  
    while (A[d] != 0) {
      y = A[d];  
      if (h->cmp(x->key, y->key) > 0) {
	tmp = x; 
	x = y; 
	y = tmp;
      }   
      fib_heap_link(h,y,x);
      A[d] = 0;
      d++;
    }
    A[d] = x;
    x = x1;                
  }
  h->min = 0;
  for(i=0; i<=D; i++) {
    if (A[i]!=0) {
      concatenate(h,A[i]);
      if (h->cmp(A[i]->key,h->min->key)<0)
	h->min=A[i];
    }
  }
}

// You need to manually free the return key of "extract_min" 
// where you call it.
void *
fibheap_extract_min(fibheap_t *h)     
{
  fibheap_node *z, *x;
  void *key;
  int i;
 
  z = h->min;
  if (z != 0) {
    x = z->child;
    if (x != 0) {
      for (i=1; i<=z->degree; i++) {
	x->p = 0;
	x = x->right;
      }
      concatenate(h,x);
    }
    z->left->right = z->right;
    z->right->left = z->left;
    if (z == z->right)
      h->min = 0;
    else {
      h->min = z->right;
      consolidate(h);
    }
    h->n --;
    z->right = z;
    z->left = z;
  }  
  if (z == 0) {
    printf("Fibheap error!\n");
    exit(1);
  }
  else { 
    key = z->key;
    free(z);
    return key;	
  }  
}

// We don't free z here.
void *
fibheap_extract_min_not_free(fibheap_t *h)  
{
  fibheap_node *z, *x;
  void *key;
  int i;
 
  z = h->min;
  if (z != 0) {
    x = z->child;
    if (x != 0) {
      for (i=1;i <=z->degree; i++) {
	x->p = 0;
	x = x->right;
      }
      concatenate(h,x);
    }
    z->left->right = z->right;
    z->right->left = z->left;
    if (z == z->right)
      h->min = 0;
    else {
      h->min = z->right;
      consolidate(h);
    }
    h->n --;
    z->right = z;
    z->left = z;
  }  
  if (z == 0) {
    printf("Fibheap error!\n");
    exit(1);
  }
  else { 
    key = z->key;
    return key;	
  }  
}

unsigned int 
fibheap_size(fibheap_t *h)
{
  return (h->n);
}

void *
fibheap_min(fibheap_t *h)
{
  return(h->min->key);
}

void 
cut(fibheap_t *h, fibheap_node *x, fibheap_node *y)
{
  if (x == x->right) 
    y->child = 0;
  else {
    y->child = x->right;
    x->left->right = x->right;
    x->right->left = x->left;
    x->left = x;
    x->right = x;
  }
  y->degree --;
  concatenate(h, x);
  x->p = 0;
  x->mark = 0;
}

void 
cascading_cut(fibheap_t *h, fibheap_node *y)
{
  fibheap_node *z;
  
  z = y->p;
  if (z != 0) {
    if (y->mark == 0)
      y->mark = 1;
    else {
      cut(h,y,z);
      cascading_cut(h,z);
    }	
  }
}


void 
fibheap_decrease_key(fibheap_t *h, fibheap_node *x, void *data)
{
  fibheap_node *y;

  if (h->cmp(data, x->key) > 0)
    printf("New key should be smaller than current key.\n");
  else {
    // If data and x->key have the same address, NOT free it.
    if (h->cmp(data, x->key) < 0)  
      free(x->key);
    x->key = data;	  	  
    y = x->p;
    if ((y!=0) && (h->cmp(x->key, y->key)<0)) {                                 
      cut(h,x,y);
      cascading_cut(h,y);
    }
    if (h->cmp(x->key, h->min->key) < 0)
      h->min = x;
  }	 
}

void 
fibheap_delete(fibheap_t *h, fibheap_node *x)
{
  void *tmp;

  fibheap_decrease_key(h, x, h->neginf); 
  tmp = fibheap_extract_min(h);
  // tmp is "neginf", NOT to free during the lifetime of heap.
  //  free(tmp);  
}

void *
fibheap_node_get_data(fibheap_node *node)
{
  return (node->key);
}

void
fibheap_check(fibheap_node *node)
{
  if ((node->right->left != node) || 
      (node->left->right != node)) {
    printf("\nstructure error\n");
    exit(1);
  }
}

void fibheap_reset(fibheap_t *h)
{
  fibheap_node *node;
    
  node = h->min;
  if (node != 0) 
    free_recursive(node, free);  
  h->min = 0;
  h->n = 0;
}

void 
fibheap_insert_node(fibheap_t *h, fibheap_node *node)
{
  node->degree = 0;
  node->p = 0;
  node->child = 0;
  node->left = node;
  node->right = node;
  node->mark = 0;
 
  concatenate(h, node);  
  if (h->cmp(node->key,h->min->key )< 0)
    h->min = node;
  h->n ++;
}
